import {
	application,
	DataSource,
	Resource,
	resource,
} from '.';

import * as v from 'valibot';
import * as dataSources from './data-sources';
import * as serializers from './serializers';

const autoIncrement = async (dataSource: DataSource) => {
	const data = await dataSource.getMultiple() as Record<string, string>[];

	const highestId = data.reduce<number>(
		(highestId, d) => (Number(d.id) > highestId ? Number(d.id) : highestId),
		-Infinity
	);

	return (highestId + 1).toString();
};


const TEXT_SERIALIZER_PAIR = {
	serialize(obj: unknown, level = 0): string {
		if (Array.isArray(obj)) {
			return obj.map((o) => this.serialize(o)).join('\n\n');
		}

		if (typeof obj === 'object') {
			if (obj !== null) {
				return Object.entries(obj)
					.map(([key, value]) => `${Array(level * 2).fill(' ').join('')}${key}: ${this.serialize(value, level + 1)}`)
					.join('\n');
			}

			return '';
		}

		if (typeof obj === 'number' && Number.isFinite(obj)) {
			return obj.toString();
		}

		if (typeof obj === 'string') {
			return obj;
		}

		return '';
	},
	deserialize: <T>(str: string) => str as T
};

const Piano = resource({
	brand: v.string()
})
	.name('Piano')
	.id('id', {
		generationStrategy: autoIncrement,
	});

// TODO implement authentication and RBAC on each resource

const User = resource({
	firstName: v.string(),
	middleName: v.string(),
	lastName: v.string(),
	bio: v.string(),
	createdAt: v.date()
})
	.name('User')
	.fullText('bio')
	.id('id', {
		generationStrategy: autoIncrement,
	});

const app = application({
	name: 'piano-service',
	dataSource: (resource: Resource) => new dataSources.jsonlFile.DataSource(resource),
})
	.contentType('application/json', serializers.applicationJson)
	.contentType('text/json', serializers.textJson)
	.contentType('text/plain', TEXT_SERIALIZER_PAIR)
	.resource(Piano)
	.resource(User);

app.createServer({
	baseUrl: '/api'
}).then((server) => {
	server.listen(3000);
});
